/********************************************************************************
 * ConnectionMaster.java
 * Copyright (c) 2002 by Stefan Buettcher
 *
 * This class manages any connections to a client machine (i.e. player)
 *
 *******************************************************************************/

import java.io.*;
import java.net.*;
import java.util.*;

public class ConnectionMaster {
  private final int CONNECTION_BACKLOG = 8;
  
  private GameServer owner;
  private Cryptor cryptor;
  private boolean running;
  private ServerSocket socket;
  private int semaphore;
  private Listener listener;
  
  private Vector connections;
  private Vector players;

  private TreeMap playerConnectionMapping;
  private TreeMap connectionPlayerMapping;
  private TreeMap namePlayerMapping;
  
  public ConnectionMaster(GameServer anOwner, int port) {
    owner = anOwner;
    running = false;
    semaphore = 1;
    connections = new Vector();
    players = new Vector();
    cryptor = new RSA();
    playerConnectionMapping = new TreeMap();
    connectionPlayerMapping = new TreeMap();
    namePlayerMapping = new TreeMap();
    try {
      socket = new ServerSocket(port, CONNECTION_BACKLOG);
      listener = new Listener(this, socket);
      listener.start();
      running = true;
      System.out.println("GameServer running. Listening on port " + port + ".");
    } catch (Exception e) {
      System.out.println("ConnectionMaster could not be started (port in use).");
      System.out.println("Exiting now.\n");
      System.exit(0);
    }
  }
  
  public boolean isRunning() {
    return running;
  }
  
  public void sendMessage(Player p, String message) {
    semaphoreP();
    Connection c = (Connection)playerConnectionMapping.get(p);
    if (c != null)
      c.sendMessage(message);
    semaphoreV();
  }
  
  public void addConnection(Socket s) {
    semaphoreP();
    Connection c = new Connection(this, s);
    connections.addElement(c);
    c.start();
    semaphoreV();
  }
  
  public void removeConnection(Connection c) {
    semaphoreP();
    Player p = (Player)connectionPlayerMapping.get(c);
    semaphoreV();
    removePlayer(p, c);
    semaphoreP();
    connections.removeElement(c);
    semaphoreV();
  }
  
  private void removePlayer(Player p, Connection c) {
    if (p != null) {
      owner.removePlayer(p);
      semaphoreP();
      players.remove(p);
      namePlayerMapping.remove(p.getName());
      playerConnectionMapping.remove(p);
      connectionPlayerMapping.remove(c);
      semaphoreV();
    }
  }
  
  public void processMessage(Connection c, String s) {
    semaphoreP();
    if ((s.length() > 9) && (s.substring(0, 9).compareTo("register ") == 0)) {
      String playerName = s.substring(9);
      playerName = playerName.replace(' ', '_');
      Player p = (Player)connectionPlayerMapping.get(c);
      if (p != null)
        c.sendMessage(Msg.messageString(Msg.ERR_ONLY_ONE_ID));
      else {
        p = (Player)namePlayerMapping.get(playerName);
        if (p != null)
          c.sendMessage(Msg.messageString(Msg.ERR_DUPL_PLAYER));
        else {
          p = new Player(playerName);
          players.addElement(p);
          namePlayerMapping.put(playerName, p);
          playerConnectionMapping.put(p, c);
          connectionPlayerMapping.put(c, p);
          owner.addPlayer(p);
          c.sendMessage(Msg.messageString(Msg.OK_REGISTERED));
          p.setConnection(c);
        }
      }
    }
    else if ((s.length() > 11) && (s.substring(0, 11).compareTo("deregister ") == 0)) {
      String playerName = s.substring(11);
      Player p = (Player)namePlayerMapping.get(playerName);
      if (p == null)
        c.sendMessage(Msg.messageString(Msg.ERR_UNREGISTERED));
      else {
        c.sendMessage(Msg.messageString(Msg.OK_DEREGISTERED));
        p.setConnection(null);
        removePlayer(p, c);
      }
    }
    else {
      Player p = (Player)connectionPlayerMapping.get(c);
      if (p == null)
        c.sendMessage(Msg.messageString(Msg.ERR_UNREGISTERED));
      else {
        semaphoreV();
        owner.processMessage(p, s);
        return;
      }
    }
    semaphoreV();
  }
  
  private void semaphoreP() {
/*    
    while (semaphore <= 0)
      try { Thread.sleep(10); }
        catch (InterruptedException ie) { }
*/
    semaphore--;
/*
    if (semaphore < 0) {
      semaphore++;
      semaphoreP();
    }
*/
  }
  
  private void semaphoreV() {
    semaphore++;
  }
  
  public Cryptor getCryptor() {
    return cryptor;
  }
  
   
}

class Listener extends Thread {
  private ConnectionMaster owner;
  private ServerSocket socket;
  
  public Listener(ConnectionMaster anOwner, ServerSocket aSocket) {
    owner = anOwner;
    socket = aSocket;
  }
  
  public void run() {
    while (true) {
      try {
        Socket s = socket.accept();
        if (s != null)
          owner.addConnection(s);
      } catch (IOException ioe) { }
    }
  }
  
}
