/********************************************************************************
 * Game.java
 * Copyright (c) 2002 by Stefan Buettcher
 *
 * This class implements the Game interface for a any type of game.
 * The createNewMatch(String, Player) method creates new instances of interface
 * Match when a player asks the Master to do so.
 *
 *******************************************************************************/

import java.util.*;

public class Game {
  private Vector matches;  // list of all matches created, of course
  private String name;  // name of this type of game, e.g. "Checkers"
  private TreeMap nameMatchMapping;  // find the matches by their names
  private ServerAI serverAI;

  public Game(String aName, GameServer aParent) {
  // just an ordinary constructor
    name = aName;
    matches = new Vector();
    nameMatchMapping = new TreeMap();
    // for some games, create an autonomous server AI that offers games
    if (name.compareTo("4Gewinnt") == 0)
      serverAI = new ServerAI("4Gewinnt_AI", ServerAI.CONNECT4_AI, this);
    if (name.compareTo("Checkers") == 0)
      serverAI = new ServerAI("Checkers_AI", ServerAI.CHECKERS_AI, this);
  }

  public String getName() {
    return name;
  }

  public void removeMatch(Match m) {
  // removes the match specified
    nameMatchMapping.remove(m.getName());
    matches.remove(m);
  }

  private Match createNewMatch(String aName, Player aPlayer) {
  // creates a new game of the usual type (specified by private name)
  // and sets its name to aName, owner to this, initiator to aPlayer
    if (name.compareTo("Checkers") == 0)
      return new Match(this, aName, aPlayer);
    if (name.compareTo("Chess") == 0)
      return new Match(this, aName, aPlayer);
    else if (name.compareTo("4Gewinnt") == 0)
      return new Match(this, aName, aPlayer);
    else  // only create matches whose rules we know
      return null;
  }

  public void processMessage(Player p, String message) {
  // process player messages
  // messages processable are: "offer GAMENAME", "join GAMENAME", "quit GAMENAME" and "gamelist"
    int index = message.indexOf(' ');
    if (index > 0) {
      String command = message.substring(0, index);
      message = message.substring(index + 1);
      if (command.compareTo("offer") == 0) {
      // if there is no game with the name specified so far, create a new one
        message.replace(' ', '_');
        if (p.getMatch() == null) {
          if (nameMatchMapping.get(message) == null) {
            Match m = createNewMatch(message, p);
            if (m != null) {
              nameMatchMapping.put(message, m);
              matches.addElement(m);
              p.sendMessage(Msg.messageString(Msg.OK_CREATED));
            }
            else
              p.sendMessage(Msg.messageString(Msg.ERR_INTERNAL_ERROR));
          }
          else
            p.sendMessage(Msg.messageString(Msg.ERR_DUPL_GAME));
        }
        else
          p.sendMessage(Msg.messageString(Msg.ERR_IN_GAME));
      }
      else if (command.compareTo("join") == 0) {
      // if the game specified exists and there a free places: join (add player)
        message.replace(' ', '_');
        if (p.getMatch() == null) {
          Match m = (Match)nameMatchMapping.get(message);
          if (m == null)
            p.sendMessage(Msg.messageString(Msg.ERR_NO_SUCH_GAME));
          else {
            if (m.addPlayer(p))
              p.sendMessage(Msg.messageString(Msg.OK_JOINED));
            else
              p.sendMessage(Msg.messageString(Msg.ERR_GAME_FULL));
          }
        }
        else
          p.sendMessage(Msg.messageString(Msg.ERR_IN_GAME));
      }
      else if (command.compareTo("quit") == 0) {
      // if Player p is a member of the game specified: kick him off!
        Match m = p.getMatch();
        if ((m == null) || (m.getName().compareTo(message) != 0))
          p.sendMessage(Msg.messageString(Msg.ERR_NOT_IN_GAME));
        else {
          m.removePlayer(p);
          p.sendMessage(Msg.messageString(Msg.OK_LEFT_GAME));
        }
      }
      else {
      // if we don't know how to handle the request: pass over to the player's current match
        if (p.getMatch() != null)
          p.getMatch().processMessage(p, command + " " + message);
        else
          p.sendMessage(Msg.messageString(Msg.ERR_ILLEGAL_FORMAT));
      }
    }
    else if (message.compareTo("gamelist") == 0) {
    // return a list of available games
      if (matches.size() == 0)
        p.sendMessage(Msg.messageString(Msg.OK_NO_GAMES));
      else {
        String result = Msg.messageString(Msg.OK_GAME_LIST);
        for (int i = 0; i < matches.size() - 1; i++)
          result += ((Match)matches.elementAt(i)).getName() + ",";
        result += ((Match)matches.elementAt(matches.size() - 1)).toString() + ".";
        p.sendMessage(result);
      }
      return;
    }
    else {
      if (p.getMatch() != null)
        p.getMatch().processMessage(p, message);
      else
        p.sendMessage(Msg.messageString(Msg.ERR_ILLEGAL_FORMAT));
    }
  }

  public void doSomeWork() {
  // this is used for periodically having the matches do their homework,
  // e.g. countdown, time-over alerts
    for (int i = 0; i < matches.size(); i++)
      ((Match)matches.elementAt(i)).doSomeWork();
  }

}
