/********************************************************************************
 * Match.java
 * Copyright (c) 2002 by Stefan Buettcher
 *
 * The interface Match defines the basic functionality of a Match.
 * Implementations are: CheckerMatch.
 *******************************************************************************/

import java.util.*;

public class Match implements Comparable {
  private static final int NO_PHASE = -1;
  private static final int LOUNGE_PHASE = 0;
  private static final int STARTUP_PHASE = 1;
  private static final int RUNNING_PHASE = 2;

  private static final int COUNTDOWN = 5;

  private String name;
  private int phase, countDown;
  private Rules rules;
  private Player initiator;
  private Vector players;
  private long lastAction;
  private Game parent;

  public Match(Game aParent, String name, Player initiator) {
    phase = NO_PHASE;
    parent = aParent;
    this.name = name;
    this.initiator = initiator;
    initiator.setMatch(this);
    initiator.setCanStart(false);
    players = new Vector();
    players.addElement(initiator);
    if (parent.getName().compareTo("Checkers") == 0)
      rules = new CheckerRules(this, players);
    else if (parent.getName().compareTo("Chess") == 0)
      rules = new ChessRules(this, players);
    else if (parent.getName().compareTo("4Gewinnt") == 0)
      rules = new FourInARowRules(this, players);
    else
      rules = null;
    phase = LOUNGE_PHASE;
  }
  
  public void processMessage(Player p, String message) {
    if (message.compareTo("playerlist") == 0) {
      String response = Msg.messageString(Msg.OK_PLAYER_LIST);
      for (int i = 0; i < players.size() - 1; i++)
        response += players.elementAt(i).toString() + ",";
      response += players.elementAt(players.size() - 1).toString() + ".";
      p.sendMessage(response);
    }
    else if (message.startsWith("chat ")) {
      p.sendMessage(Msg.messageString(Msg.OK));
      sendNewsToAll("Chat from " + p.toString() + ": " + message.substring(5));
    }
    else if (message.compareTo("canstart=true") == 0) {
      p.setCanStart(true);
      p.sendMessage(Msg.messageString(Msg.OK_READY));
    }
    else if (message.compareTo("canstart=false") == 0) {
      p.setCanStart(true);
      p.sendMessage(Msg.messageString(Msg.OK_NOT_READY));
    }
    else if (message.compareTo("getboard") == 0)
      p.sendMessage(Msg.messageString(Msg.OK_BOARD) + rules.getBoard());
    else if (message.startsWith("move ")) {
      String move = message.substring(5);
      if (rules.isLegalMove(move, p)) {
        p.sendMessage(Msg.messageString(Msg.OK_LEGAL_MOVE));
        performMove(p, move);
      }
      else
        if (rules.mayMove(p))
          p.sendMessage(Msg.messageString(Msg.ERR_ILLEGAL_MOVE));
        else
          p.sendMessage(Msg.messageString(Msg.ERR_NOT_YOUR_TURN));
    }
    else if (message.compareTo("timeleft") == 0)
      p.sendMessage(Msg.messageString(Msg.OK_TIME_LEFT) + rules.getTimeLeft());
    else if (message.compareTo("whoseturn") == 0)
      p.sendMessage(Msg.messageString(Msg.OK) + rules.getWhoseTurn());
    else
      p.sendMessage(Msg.messageString(Msg.ERR_ILLEGAL_FORMAT));
    
    if ((phase == LOUNGE_PHASE) && (canStart())) {
      mixPlayersRandomly();
      phase = STARTUP_PHASE;
      countDown = COUNTDOWN;
      lastAction = System.currentTimeMillis();
      sendNewsToAll("Countdown: " + countDown + " seconds.");
    }
  }
  
  private void mixPlayersRandomly() {
    int i;
    Random r = new Random();
    Vector p2 = new Vector();
    while (players.size() > 0) {
      i = r.nextInt(players.size());
      Player p = (Player)players.elementAt(i);
      players.remove(i);
      p2.addElement(p);
    }
    for (i = 0; i < p2.size(); i++)
      players.addElement(p2.elementAt(i));
  }

  private void performMove(Player p, String move) {
    if (move != null) {
      sendNewsToAll(p.toString() + " has moved: " + move);
      rules.doMove(move);
    }
    sendNewsToAll("board=" + rules.getBoard());
    int winner = rules.isEndState();
    if (winner >= 0) {
      if (winner > players.size()) {
        sendNewsToAll("Draw game.");
      }
      else {
        for (int i = 0; i < players.size(); i++)
          if (i != winner)
            ((Player)players.elementAt(i)).sendMessage("[NEWS] " + p.toString() + " has won.");
        p.sendMessage("[NEWS] You have won.");
      }
      phase = LOUNGE_PHASE;
      while (players.size() > 0)
        removePlayer((Player)players.elementAt(0));
      return;
    }
    else {
      rules.nextTurn();
      rules.getWhoseTurn().sendMessage("[NEWS] Please move. Time left: " + rules.getTimeOut());
    }
  }

  public void doSomeWork() {
    if (phase == STARTUP_PHASE) {
      long now = System.currentTimeMillis();
      if (lastAction < now - 1000) {
        countDown--;
        sendNewsToAll("Countdown: " + countDown + " seconds.");
        lastAction = now;
        if (countDown == 0) {
          phase = RUNNING_PHASE;
          for (int i = 0; i < players.size(); i++)
            ((Player)players.elementAt(i)).sendMessage("[NEWS] You are " +
              rules.getPlayerPieces((Player)players.elementAt(i)) + ".");
          performMove(null, null);
        }
      }
    }
    else if (phase == RUNNING_PHASE) {
      long timeLeft = rules.getTimeLeft();
      if (timeLeft < 0) {
        Player p = rules.getWhoseTurn();
        p.sendMessage("[NEWS] Time out.");
        performMove(p, rules.getAnyMove());
      }
    }
  }
  
  public void sendNewsToAll(String news) {
    for (int i = 0; i < players.size(); i++)
      ((Player)players.elementAt(i)).sendMessage("[NEWS] " + news);
  }

  public void setName(String aName) {
    name = aName;
  }
  
  public String getName() {
    return name;
  }
  
  private boolean canStart() {
    if (phase == LOUNGE_PHASE) {
      if (players.size() == rules.getMinPlayers()) {
        for (int i = 0; i < players.size(); i++)
          if (!((Player)players.elementAt(i)).getCanStart())
            return false;
        return true;
      }
    }
    return false;
  }

  public boolean addPlayer(Player p) {
    if ((players.size() < rules.getMaxPlayers()) && (phase == LOUNGE_PHASE)) {
      sendNewsToAll("Player " + p.getName() + " entered the game.");
      players.addElement(p);
      p.setMatch(this);
      p.setCanStart(false);
      return true;
    }
    else
      return false;
  }

  public void removePlayer(Player p) {
    players.remove(p);
    p.setMatch(null);
    sendNewsToAll(p.getName() + " left the game.");
    if (phase == STARTUP_PHASE) {
      phase = LOUNGE_PHASE;
      sendNewsToAll("Countdown stopped.");
    }
    if (players.size() == 0)
      parent.removeMatch(this);
    else if ((players.size() == 1) && (phase == RUNNING_PHASE)) {
      Player p2 = (Player)players.elementAt(0);
      sendNewsToAll("You have won.");
      if (players.size() > 0)
        removePlayer(p2);
    }
    phase = LOUNGE_PHASE;
  }
  
  public boolean getRunning() {
    return (phase == RUNNING_PHASE);
  }
  
  public String toString() {
    return name + "(" + initiator.getName() + ")";
  }
  
  public int compareTo(Object o) {
    return name.compareTo(o);
  }

}


interface Rules {
  
  public boolean doMove(String move);
  public String getAnyMove();
  public String getBoard();
  public int getMaxPlayers();
  public int getMinPlayers();
  public String getPlayerPieces(Player p);
  public long getTimeLeft();
  public long getTimeOut();
  public Player getWhoseTurn();
  public int isEndState();
  public boolean isLegalMove(String move, Player p);
  public boolean mayMove(Player p);
  public void nextTurn();

}


class CheckerRules implements Rules {
  private static final int TIME_PER_TURN = 40000;

  private char board[][];
  private Match owner;
  private int whoseTurn;
  private Vector players;
  private long moveTimeStart;
  private int turnCount;
  private Random random;

  public CheckerRules(Match owner, Vector players) {
    int j, k;
    this.owner = owner;
    this.players = players;
    board = new char[8][8];
    whoseTurn = 1;   // white begins
    turnCount = -1;  // indicates start of the game
    // now, setup the initial board state
    for (j = 0; j < 8; j++)
      for (k = 0; k < 8; k++)
        board[j][k] = '.';
    for (j = 0; j < 4; j++) {
      board[j * 2][0] = 'w';
      board[j * 2 + 1][1] = 'w';
      board[j * 2][2] = 'w';
      board[j * 2 + 1][5] = 'b';
      board[j * 2][6] = 'b';
      board[j * 2 + 1][7] = 'b';
    }
    random = new Random();
  }

  public String getPlayerPieces(Player p) {
    if (p == (Player)players.elementAt(0))
      return "w";
    else
      return "b";
  }

  public long getTimeOut() {
    return TIME_PER_TURN;
  }

  public boolean mayMove(Player p) {
    if (p == (Player)players.elementAt(whoseTurn))
      return true;
    else
      return false;
  }

  public long getTimeLeft() {
    return moveTimeStart + TIME_PER_TURN - System.currentTimeMillis();
  }

  private void doMoveInternal(String move) {
    move = move.toLowerCase();
    if (move.length() == 4) {
      int fromX = move.charAt(0) - 'a';
      int fromY = move.charAt(1) - '1';
      int toX = move.charAt(2) - 'a';
      int toY = move.charAt(3) - '1';
      if ((toY == fromY + 1) || (toY == fromY - 1)) {
        board[toX][toY] = board[fromX][fromY];
        board[fromX][fromY] = '.';
      }
      else if ((toY == fromY + 2) || (toY == fromY - 2)) {
        board[toX][toY] = board[fromX][fromY];
        board[(toX + fromX) / 2][(toY + fromY) / 2] = '.';
        board[fromX][fromY] = '.';
      }
      if ((board[toX][toY] == 'w') && (toY == 7))
        board[toX][toY] = 'W';
      else if ((board[toX][toY] == 'b') && (toY == 0))
        board[toX][toY] = 'B';
    }
    else {
      doMoveInternal(move.substring(0, 4));
      doMoveInternal(move.substring(2));
    }
  }

  public boolean doMove(String move) {
    doMoveInternal(move);
    return true;
  }

  public int getMaxPlayers() {
    return 2;
  }

  public int getMinPlayers() {
    return 2;
  }

  private int isEndStateInternal(char aBoard[][]) {
    int w = 0;
    int b = 0;
    for (int y = 0; y < 8; y++)
      for (int x = 0; x < 8; x++) {
        if ((aBoard[x][y] == 'w') || (aBoard[x][y] == 'W')) w++;
        if ((aBoard[x][y] == 'b') || (aBoard[x][y] == 'B')) b++;
      }
    if (w == 0) return 1;
    if (b == 0) return 0;
    return -1;
  }

  public int isEndState() {
    return isEndStateInternal(board);
  }

  public String getBoard() {
    String result = "";
    for (int y = 0; y < 8; y++)
      for (int x = 0; x < 8; x++)
        result += "" + board[x][y];
    return result;
  }

  public boolean isLegalMove(String move, Player p) {
    if (p == (Player)players.elementAt(whoseTurn))
      return isLegalMoveInternal(duplicateBoard(board), move, whoseTurn);
    else
      return false;
  }

  private char[][] duplicateBoard(char board[][]) {
    char result[][] = new char[8][8];
    for (int x = 0; x < 8; x++)
      for (int y = 0; y < 8; y++)
        result[x][y] = board[x][y];
    return result;
  }

  public void nextTurn() {
    moveTimeStart = System.currentTimeMillis();
    whoseTurn = 1 - whoseTurn;
    turnCount++;
  }

  private boolean isLegalMoveInternal(char aBoard[][], String move, int whoseTurn) {
    int fromX, fromY, toX, toY, midX, midY;
    char enemy1, enemy2;
    boolean mustAllBeWar;  // indicates if this is a multi-move, killing enemy troups

    if (move.length() < 4)
      return false;

    move = move.toLowerCase();

    if (move.length() > 4)
      mustAllBeWar = true;
    else
      mustAllBeWar = false;

    while (move.length() >= 4) {
      fromX = move.charAt(0) - 'a';
      fromY = move.charAt(1) - '1';
      toX = move.charAt(2) - 'a';
      toY = move.charAt(3) - '1';
      midX = (fromX + toX) / 2;
      midY = (fromY + toY) / 2;
      move = move.substring(2);

      // please do not move out of the board
      if ((fromX < 0) || (fromX > 7) || (fromY < 0) || (fromY > 7) ||
          (toX < 0) || (toX > 7) || (toY < 0) || (toY > 7))
        return false;

      // you may only move your own pieces
      if ((whoseTurn == 0) && (aBoard[fromX][fromY] != 'w') && (aBoard[fromX][fromY] != 'W'))
        return false;
      if ((whoseTurn == 1) && (aBoard[fromX][fromY] != 'b') && (aBoard[fromX][fromY] != 'B'))
        return false;

      switch (aBoard[fromX][fromY]) {
        case '.': return false;
        case 'w':
          if ((toY == fromY + 1) && ((toX == fromX - 1) || (toX == fromX + 1)) &&
              (!mustAllBeWar) && (aBoard[toX][toY] == '.')) {
            aBoard[toX][toY] = aBoard[fromX][fromY];
            aBoard[fromX][fromY] = '.';
          }
          else if ((toY == fromY + 2) && ((toX == fromX - 2) || (toX == fromX + 2)) &&
                   (aBoard[toX][toY] == '.') &&
                   ((aBoard[midX][midY] == 'b') || (aBoard[midX][midY] == 'B'))) {
            aBoard[toX][toY] = aBoard[fromX][fromY];
            aBoard[fromX][fromY] = '.';
            aBoard[midX][midY] = '.';
          }
          else
            return false;
          break;
        case 'b':
          if ((toY == fromY - 1) && ((toX == fromX - 1) || (toX == fromX + 1)) &&
              (!mustAllBeWar) && (aBoard[toX][toY] == '.')) {
            aBoard[toX][toY] = aBoard[fromX][fromY];
            aBoard[fromX][fromY] = '.';
          }
          else if ((toY == fromY - 2) && ((toX == fromX - 2) || (toX == fromX + 2)) &&
                   (aBoard[toX][toY] == '.') &&
                   ((aBoard[midX][midY] == 'w') || (aBoard[midX][midY] == 'W'))) {
            aBoard[toX][toY] = aBoard[fromX][fromY];
            aBoard[fromX][fromY] = '.';
            aBoard[midX][midY] = '.';
          }
          else
            return false;
          break;
        case 'W': case 'B':
          if (aBoard[fromX][fromY] == 'W') {
            enemy1 = 'b'; enemy2 = 'B';
          }
          else {
            enemy1 = 'w'; enemy2 = 'W';
          }
          if (((toY == fromY + 1) || (toY == fromY - 1)) && ((toX == fromX - 1) || (toX == fromX + 1)) &&
              (!mustAllBeWar) && (aBoard[toX][toY] == '.')) {
            aBoard[toX][toY] = aBoard[fromX][fromY];
            aBoard[fromX][fromY] = '.';
          }
          else if (((toY == fromY - 2) || (toY == fromY + 2)) && ((toX == fromX - 2) || (toX == fromX + 2)) &&
                   (aBoard[toX][toY] == '.') &&
                   ((aBoard[midX][midY] == enemy1) || (aBoard[midX][midY] == enemy2))) {
            aBoard[toX][toY] = aBoard[fromX][fromY];
            aBoard[fromX][fromY] = '.';
            aBoard[midX][midY] = '.';
          }
          else
            return false;
          break;
        default: return false;
      }
    }

    if (move.length() != 2)
      return false;  // wrong format, companero!

    return true;
  }
  
  public String getAnyMove() {
    Vector moves = new Vector();
    char piece;
    int direction, startFrom;
    if (whoseTurn == 0) {
      piece = 'w';
      direction = +1;
      startFrom = 0;
    }
    else {
      piece = 'b';
      direction = -1;
      startFrom = 7;
    }
    for (char x = 0; x < 8; x++)
      for (char y = 0; y < 7; y++) {
        int tempY = startFrom + y * direction; 
        if (board[x][tempY] == piece) {
          if ((x > 0) && (board[x - 1][tempY + direction] == '.'))
            moves.add("" + (char)(x + 'a') + "" + (char)(tempY + '1') + "" + 
                      (char)(x - 1 + 'a') + "" + (char)(tempY + direction + '1'));
          if ((x < 7) && (board[x + 1][tempY + direction] == '.'))
            moves.add("" + (char)(x + 'a') + "" + (char)(tempY + '1') + "" +
                      (char)(x + 1 + 'a') + "" + (char)(tempY + direction + '1'));
        }
      }
    String result = (String)moves.elementAt(random.nextInt(moves.size()));
    return result;
  }
  
  public Player getWhoseTurn() {
    return (Player)players.elementAt(whoseTurn);
  }
  
}


class ChessRules implements Rules {
  private static final int TIME_PER_TURN = 150000;
  
  private char board[][];
  private Match owner;
  private int whoseTurn;
  private Vector players;
  private long moveTimeStart;
  private int turnCount;
  private Random random;
  private boolean mayDoRochade[];

  public ChessRules(Match owner, Vector players) {
    this.owner = owner;
    this.players = players;
    board = new char[8][8];
    whoseTurn = 1;   // white begins
    turnCount = -1;  // indicates start of the game
    // now, setup the initial board state
    String boardStr = "RKBQCBKR" + "PPPPPPPP" + "........" + "........" +
                      "........" + "........" + "pppppppp" + "rkbqcbkr";
    for (int j = 0; j < 64; j++)
      board[j%8][j/8] = boardStr.charAt(j);
    random = new Random();
    mayDoRochade = new boolean[2];
    mayDoRochade[0] = mayDoRochade[1] = true;
  }
  
  public String getPlayerPieces(Player p) {
    if (p == (Player)players.elementAt(0))
      return "X";
    else
      return "x";
  }

  public long getTimeOut() {
    return TIME_PER_TURN;
  }

  public boolean mayMove(Player p) {
    if (p == (Player)players.elementAt(whoseTurn))
      return true;
    else
      return false;
  }

  public long getTimeLeft() {
    return moveTimeStart + TIME_PER_TURN - System.currentTimeMillis();
  }

  private void doMoveInternal(char board[][], String move) {
    move = move.toLowerCase();
    if (move.length() == 4) {
      int fromX = move.charAt(0) - 'a';
      int fromY = move.charAt(1) - '1';
      int toX = move.charAt(2) - 'a';
      int toY = move.charAt(3) - '1';
      doMoveInternal(board, fromX, fromY, toX, toY);
    }
  }

  private void doMoveInternal(char board[][], int fromX, int fromY, int toX, int toY) {
    if ((board[fromX][fromY] == 'c') || (board[fromX][fromY] == 'C'))
      mayDoRochade[whoseTurn] = false;
    if (((board[fromX][fromY] == 'c') || (board[fromX][fromY] == 'C')) &&
        ((toX > fromX + 1) || (toX < fromX - 1))) {
      board[toX][toY] = board[fromX][fromY];
      board[fromX][fromY] = '.';
      if (toX > fromX) {
        board[toX - 1][toY] = board[7][toY];
        board[7][toY] = '.';
      }
      else {
        board[toX + 1][toY] = board[0][toY];
        board[0][toY] = '.';
      }
    }
    else {
      board[toX][toY] = board[fromX][fromY];
      board[fromX][fromY] = '.';
    }
  }

  public boolean doMove(String move) {
    doMoveInternal(board, move);
    return true;
  }

  public int getMaxPlayers() {
    return 2;
  }

  public int getMinPlayers() {
    return 2;
  }

  private boolean isCheck() {
    return false;
  }

  private int isEndStateInternal(char board[][], int whoseTurn) {
    int kingCount = 0;
    for (int x = 0; x < 8; x++)
      for (int y = 0; y < 8; y++) {
        if (board[x][y] == 'C')
          kingCount++;
        if (board[x][y] == 'c')
          kingCount--;
      }
    if (kingCount > 0) return 0;
    if (kingCount < 0) return 1;

    String move = getAnyMoveInternal(board, 1 - whoseTurn);

    if (move == null) {
      if (mayKillTheKing(board, whoseTurn))
        return 1 - whoseTurn;
      else
        return 23;
    }

    return -1;
  }

  public int isEndState() {
    return isEndStateInternal(board, whoseTurn);
  }

  public String getBoard() {
    String result = "";
    for (int y = 0; y < 8; y++)
      for (int x = 0; x < 8; x++)
        result += "" + board[x][y];
    return result;
  }

  public boolean isLegalMove(String move, Player p) {
    if (p == (Player)players.elementAt(whoseTurn))
      return isLegalMoveInternal(board, move, whoseTurn);
    else
      return false;
  }

  private char[][] duplicateBoard(char board[][]) {
    char result[][] = new char[8][8];
    for (int x = 0; x < 8; x++)
      for (int y = 0; y < 8; y++)
        result[x][y] = board[x][y];
    return result;
  }

  public void nextTurn() {
    moveTimeStart = System.currentTimeMillis();
    whoseTurn = 1 - whoseTurn;
    turnCount++;
  }

  private int abs(int i) {
    if (i >= 0) return i;
      else return -i;
  }

  private boolean checkForFree(char board[][], int fromX, int fromY, int toX, int toY) {
    int incX, incY;
    if ((fromX == toX) && (fromY != toY)) {
      if (toY < fromY) incY = -1;
        else incY = +1;
      incX = 0;
    }
    else if ((fromX != toX) && (fromY == toY)) {
      if (toX < fromX) incX = -1;
        else incX = +1;
      incY = 0;
    }
    else {
      if (abs(toX - fromX) != abs(toY - fromY))
        return false;
      if (toX < fromX) incX = -1;
        else incX = +1;
      if (toY < fromY) incY = -1;
        else incY = +1;
    }
    int x = fromX + incX;
    int y = fromY + incY;
    while ((x != toX) || (y != toY)) {
      if (board[x][y] != '.')
        return false;
      x += incX;
      y += incY;
    }
    return true;
  }

  private boolean isLegalMoveInternal(char board[][], String move, int whoseTurn) {
    int fromX, fromY, toX, toY;
    if ((move == null) || (move.length() < 4))
      return false;
    else
      move = move.toLowerCase();
    fromX = move.charAt(0) - 'a';
    fromY = move.charAt(1) - '1';
    toX = move.charAt(2) - 'a';
    toY = move.charAt(3) - '1';
    return isLegalMoveInternal(board, fromX, fromY, toX, toY, whoseTurn);
  }

  private boolean isLegalMoveInternal(char board[][], int fromX, int fromY, int toX, int toY, int whoseTurn) {
    if (!isTechnicallyCorrectMove(fromX, fromY, toX, toY, board, whoseTurn))
      return false;
    char newBoard[][] = duplicateBoard(board);
    doMoveInternal(newBoard, fromX, fromY, toX, toY);
    return (!mayKillTheKing(newBoard, 1 - whoseTurn));
  }

  private boolean isTechnicallyCorrectMove(int fromX, int fromY, int toX, int toY, char board[][], int whoseTurn) {
    if ((fromX < 0) || (fromX > 7) || (fromY < 0) || (fromY > 7) ||
        (toX < 0) || (toX > 7) || (toY < 0) || (toY > 7))
      return false;

    if ((fromX == toX) && (fromY == toY))
      return false;

    char c = board[fromX][fromY];

    if (whoseTurn == 0) {  // white
      if ((c < 'A') || (c > 'Z'))
        return false;
      if ((board[toX][toY] >= 'A') && (board[toX][toY] <= 'Z'))
        return false;
    }
    if (whoseTurn == 1) {  // black
      if ((c < 'a') || (c > 'z'))
        return false;
      if ((board[toX][toY] >= 'a') && (board[toX][toY] <= 'z'))
        return false;
    }

    // first, check the technical constraints
    if ((c == 'c') || (c == 'C')) { // Chief = King
      if (((toX == fromX - 2) || (toX == fromX + 2)) &&
          (mayDoRochade[whoseTurn]) && (toY == fromY)) {
        if ((toX > fromX) && ((board[fromX + 1][fromY] != '.') || (board[fromX + 2][fromY] != '.')))
          return false;
        if ((toX < fromX) && ((board[fromX - 1][fromY] != '.') || (board[fromX - 2][fromY] != '.')))
          return false;
      }
      else if ((toX < fromX - 1) || (toX > fromX + 1) ||
          (toY < fromY - 1) || (toY > fromY + 1))
        return false;
    }

    if (c == 'P') { // white pawn
      if (toX != fromX) {  // catcher in the rye
        if (toY != fromY + 1)
          return false;  // only one step
        if ((toX != fromX - 1) && (toX != fromX + 1))
          return false;  // only one step
        if ((board[toX][toY] < 'a') || (board[toX][toY] > 'z'))
          return false;  // may only catch enemy pieces
      }
      else {
        if (board[toX][toY] != '.')
          return false;  // only move to empty fields
        if (toY == fromY + 1) {
        }
        else if (toY == fromY + 2) {
          if ((fromY != 1) || (board[fromX][fromY + 1] != '.'))
            return false;
        }
        else
          return false;
      }
    }
    if (c == 'p') { // black pawn
      if (toX != fromX) {  // catcher in the rye
        if (toY != fromY - 1)
          return false;  // only one step
        if ((toX != fromX - 1) && (toX != fromX + 1))
          return false;  // only one step
        if ((board[toX][toY] < 'A') || (board[toX][toY] > 'Z'))
          return false;  // may only catch enemy pieces
      }
      else {
        if (board[toX][toY] != '.')
          return false;  // only move to empty fields
        if (toY == fromY - 1) {
        }
        else if (toY == fromY - 2) {
          if ((fromY != 6) || (board[fromX][fromY - 1] != '.'))
            return false;
        }
        else
          return false;
      }

    }

    if ((c == 'K') || (c == 'k')) { // Knight
      int dx = abs(toX - fromX);
      int dy = abs(toY - fromY);
      if (((dx != 2) || (dy != 1)) && ((dx != 1) || (dy != 2)))
        return false;
    }

    if ((c == 'B') || (c == 'b') || (c == 'R') || (c == 'r') ||
        (c == 'Q') || (c == 'q')) {
      if (!checkForFree(board, fromX, fromY, toX, toY))
        return false;
      if ((c == 'B') || (c == 'b'))
        if ((toX == fromX) || (toY == fromY))
          return false;
      if ((c == 'R') || (c == 'r'))
        if ((toX != fromX) && (toY != fromY))
          return false;
    }

    return true;
  }  // end of isTechnicallyCorrect(int, int, int, int, char[][], int)

  private boolean mayKillTheKing(char board[][], int whoseTurn) {
    char toLookFor;
    int x, y, kingX, kingY;
    if (whoseTurn == 0)
      toLookFor = 'c';
    else
      toLookFor = 'C';
    kingX = kingY = 0;
    for (x = 0; x < 8; x++)
      for (y = 0; y < 8; y++)
        if (board[x][y] == toLookFor) {
          kingX = x;
          kingY = y;
        }
    for (x = 0; x < 8; x++)
      for (y = 0; y < 8; y++)
        if (isTechnicallyCorrectMove(x, y, kingX, kingY, board, whoseTurn))
          return true;
    return false;
  }  // end of mayKillTheKing(char[][], int)

  public String getAnyMove() {
    return getAnyMoveInternal(board, whoseTurn);
  }

  private String getAnyMoveInternal(char board[][], int whoseTurn) {
    Vector moves = new Vector();
    int x1, y1, x2, y2;
    for (x1 = 0; x1 < 8; x1++)
      for (y1 = 0; y1 < 8; y1++)
        for (x2 = 0; x2 < 8; x2++)
          for (y2 = 0; y2 < 8; y2++)
            if (isLegalMoveInternal(board, x1, y1, x2, y2, whoseTurn))
              moves.add("" + (char)('a' + x1) + "" + (char)('1' + y1) + "" + (char)('a' + x2) + "" + (char)('1' + y2));
    if (moves.size() == 0)
      return null;
    else
      return (String)moves.elementAt(random.nextInt(moves.size()));
  }

  public Player getWhoseTurn() {
    return (Player)players.elementAt(whoseTurn);
  }

}


class FourInARowRules implements Rules {
  private static final int TIME_PER_TURN = 30000;

  private char board[][];
  private Match owner;
  private int whoseTurn;
  private Vector players;
  private long moveTimeStart;
  private int turnCount;
  private Random random;
  
  public FourInARowRules(Match owner, Vector players) {
    int j, k;
    this.owner = owner;
    this.players = players;
    board = new char[7][6];
    whoseTurn = 1;   // white begins
    turnCount = -1;  // indicates start of the game
    // now, setup the initial board state
    for (j = 0; j < 7; j++)
      for (k = 0; k < 6; k++)
        board[j][k] = '.';
    random = new Random();
  }
  
  public String getPlayerPieces(Player p) {
    if (p == (Player)players.elementAt(0))
      return "w";
    else
      return "b";
  }
  
  public long getTimeOut() {
    return TIME_PER_TURN;
  }
  
  public boolean mayMove(Player p) {
    if (p == (Player)players.elementAt(whoseTurn))
      return true;
    else
      return false;
  }
  
  public long getTimeLeft() {
    return moveTimeStart + TIME_PER_TURN - System.currentTimeMillis();
  }
  
  public boolean doMove(String move) {
    if ((move == null) || (move.length() != 1))
      return false;
    int x = (char)(move.charAt(0) - '0');
    if ((x < 0) || (x >= 7))
      return false;
    int y = 5;
    while ((y >= 0) && (board[x][y] != '.'))
      y--;
    if (y >= 0) {
      if (whoseTurn == 0)
        board[x][y] = 'w';
      else
        board[x][y] = 'b';
      return true;
    }
    else
      return false;
  }

  public int getMaxPlayers() {
    return 2;
  }
  
  public int getMinPlayers() {
    return 2;
  }
  
  private boolean row(int x, int y, int count, int xInc, int yInc, char piece) {
  // returns true if there is a row in the direction indicated by (xInc, yInc)
  // of length >= count of pieces of type piece
    if (count <= 0)
      return true;
    if ((x < 0) || (x >= 7) || (y < 0) || (y >= 6))
      return false;
    return (board[x][y] == piece) &
           (row(x + xInc, y + yInc, count - 1, xInc, yInc, piece));
  }

  public int isEndState() {
  // returns true if the current board state is a final state
    int x, y, winner;
    for (y = 0; y < 6; y++)
      for (x = 0; x < 7; x++) {
        char c = board[x][y];
        if (c != '.') {
          if (c == 'w') winner = 0;
            else winner = 1;
          if (row(x, y, 4, 1, 0, c))
            return winner;
          if (row(x, y, 4, 0, 1, c))
            return winner;
          if (row(x, y, 4, 1, 1, c))
            return winner;
          if (row(x, y, 4, -1, 1, c))
            return winner;
        }
      }
    return -1;
  }

  public String getBoard() {
    String result = "";
    for (int y = 0; y < 6; y++)
      for (int x = 0; x < 7; x++)
        result += "" + board[x][y];
    return result;
  }
  
  public boolean isLegalMove(String move, Player p) {
    if (p == (Player)players.elementAt(whoseTurn)) {
      if ((move == null) || (move.length() != 1)) return false;
      int x = (char)(move.charAt(0) - '0');
      if ((x < 0) || (x >= 7)) return false;
      int y = 5;
      while ((y >= 0) && (board[x][y] != '.')) y--;
      return (y >= 0);
    }
    else
      return false;
  }
  
  private char[][] duplicateBoard(char board[][]) {
    char result[][] = new char[7][6];
    for (int x = 0; x < 7; x++)
      for (int y = 0; y < 6; y++)
        result[x][y] = board[x][y];
    return result;
  }
  
  public void nextTurn() {
    moveTimeStart = System.currentTimeMillis();
    whoseTurn = 1 - whoseTurn;
    turnCount++;
  }
  
  public String getAnyMove() {
    Vector moves = new Vector();
    char piece;
    int direction, startFrom;
    for (char x = 0; x < 7; x++)
      if (board[x][0] == '.')
        moves.add("" + (char)((char)x + '0'));
    String result = (String)moves.elementAt(random.nextInt(moves.size()));
    return result;
  }

  public Player getWhoseTurn() {
    return (Player)players.elementAt(whoseTurn);
  }

}
