/********************************************************************************
 * RSA.java
 * Copyright (c) 2002 by Stefan Buettcher
 *
 * The class RSA is the RSA implementation of the public-key-based crypto
 * interface Cryptor.
 *
 *******************************************************************************/

import java.math.*;
import java.util.*;

public class RSA implements Cryptor {
  private static final int BIT_LENGTH = 128;
  private static final int CERTAINTY = 50;
  private static final int PIECE_LENGTH = 6;

  private BigInteger modulus, privateKey, publicKey;

  public RSA(BigInteger p, BigInteger q, BigInteger priv) {
  // create a new RSA instance with modulus n = p*q and private key priv
    modulus = p.multiply(q);
    p = p.subtract(BigInteger.valueOf(1));
    q = q.subtract(BigInteger.valueOf(1));
    privateKey = priv;
    publicKey = privateKey.modInverse(p.multiply(q));
  }

  public RSA(BigInteger n, BigInteger pub) {
  // create a new RSA instance with modulus n and public key pub. this instance has got no private key!
    modulus = n;
    privateKey = null;
    publicKey = pub;
  }

  public RSA() {
  // create a new RSA instance with private and public key and random parameters
    Random r = new Random(System.currentTimeMillis());
    BigInteger p = new BigInteger(BIT_LENGTH/2, CERTAINTY, r);
    BigInteger q = new BigInteger(BIT_LENGTH/2, CERTAINTY, r);
    modulus = p.multiply(q);
    p = p.subtract(BigInteger.valueOf(1));
    q = q.subtract(BigInteger.valueOf(1));
    privateKey = new BigInteger(BIT_LENGTH/3, CERTAINTY, r);
    publicKey = privateKey.modInverse(p.multiply(q));
    p = null; q = null;
    System.gc();
  }

  public RSA(String importString) {
  // if the argument String has been created using the exportPublicKey() method,
  // this constructor creates a new instance with modulus and public key
    privateKey = null;
    if (importString.substring(0, 3).compareTo("RSA") == 0) {
      importString = importString.substring(3);
      int i = importString.indexOf("#");
      if (i > 0) {
        modulus = stringToNumber(importString.substring(0, i));
        publicKey = stringToNumber(importString.substring(i + 1));
        return;
      }
    }
    modulus = publicKey = null;
  }

  public BigInteger getModulus() {
    return modulus;
  }

  public BigInteger getPublicKey() {
    return publicKey;
  }

  public String exportPublicKey() {
  // exports the modulus and public key so that they can be read by the
  // RSA(String) constructor
    if (publicKey == null)
      return "";
    return "RSA" + numberToString(modulus) + "#" + numberToString(publicKey);
  }

  private int byteToInt(byte b) {
    if (b >= 0)
      return b;
    else
      return ((int)b) + 256;
  }
  
  private byte intToByte(int i) {
    if ((i >= 0) && (i < 128))
      return (byte)i;
    else
      return (byte)(i - 256);
  }
  
  private String encodeString(String text) {
  // transforms an 8-bit string to a 6-bit string
    char mod = (char)((char)text.length() % 3 + 'A');
    String result = "";
    while (text.length() % 3 != 0)
      text += "_";
    while (text.length() > 0) {
      byte bytes[] = text.substring(0, 3).getBytes();
      text = text.substring(3);
      int value = byteToInt(bytes[0]) + byteToInt(bytes[1]) * 256 +
                  byteToInt(bytes[2]) * 65536;
      byte bytes2[] = new byte[4];
      bytes2[0] = (byte)(value % 64 + 48); value /= 64;
      bytes2[1] = (byte)(value % 64 + 48); value /= 64;
      bytes2[2] = (byte)(value % 64 + 48); value /= 64;
      bytes2[3] = (byte)(value % 64 + 48);
      result += new String(bytes2);
    }
    return "" + mod + result;
  } // end of encodeString(String)

  private String decodeString(String text) {
  // retransforms the 6-bit string to an 8-bit string
    int mod = (char)(text.charAt(0) - 'A');
    text = text.substring(1);
    String result = "";
    while (text.length() > 0) {
      byte bytes[] = text.substring(0, 4).getBytes();
      text = text.substring(4);
      int value = (int)(bytes[0] - 48) + (int)(bytes[1] - 48) * 64 + (int)(bytes[2] - 48) * 64*64 +
                  (int)(bytes[3] - 48) * 64*64*64;
      byte bytes2[] = new byte[3];
      bytes2[0] = intToByte(value % 256); value /= 256;
      bytes2[1] = intToByte(value % 256); value /= 256;
      bytes2[2] = intToByte(value % 256);
      result += new String(bytes2);
    }
    mod = (3 - mod) % 3;
    return result.substring(0, result.length() - mod);
  } // end of decodeString(String)

  private BigInteger stringToNumber(String s) {
  // converts a string to a number
    BigInteger result = BigInteger.ZERO;
    for (int i = 0; i < s.length(); i++) {
      result = result.multiply(BigInteger.valueOf(64));
      result = result.add(BigInteger.valueOf(s.charAt(i) - 48));
    }
    return result;
  }

  private String numberToString(BigInteger i) {
  // converts a number to a string
    BigInteger mod = BigInteger.valueOf(64);
    String result = "";
    while (i.compareTo(BigInteger.ZERO) > 0) {
      char c = (char)((char)i.mod(mod).intValue() + 48);
      result = "" + c + result;
      i = i.divide(mod);
    }
    return result;
  }

  private Vector splitString(String text) {
  // since the modulus is finite, we can only encrypt string up to a certain length
  // splitString splits strings that are bigger into several sub-strings
    Vector result = new Vector();
    while (text.length() > PIECE_LENGTH) {
      String piece = text.substring(0, PIECE_LENGTH);
      text = text.substring(PIECE_LENGTH);
      result.addElement(text);
    }
    if (text.length() > 0)
      result.addElement(text);
    return result;
  }

  private BigInteger encryptNumber(BigInteger x) {
  // RSA encrypt function
    return x.modPow(publicKey, modulus);
  }

  private BigInteger decryptNumber(BigInteger y) {
  // RSA decrypt function
    return y.modPow(privateKey, modulus);
  }

  private String encryptShort(String plainText) {
  // encrypts a short string
    if (publicKey == null) return null;
    return numberToString(encryptNumber(stringToNumber(encodeString(plainText))));
  }

  private String decryptShort(String cipherText) {
  // decrypts a short string
    if (privateKey == null) return null;
    return decodeString(numberToString(decryptNumber(stringToNumber(cipherText))));
  }

  private String hashValue(String text) {
  // creates a funny hash value
    return "" + text.hashCode();
  }

  public String encrypt(String plainText) {
  // encrypts arbitrary strings
    String result = "";
    String dummy;
    while (plainText.length() > 0) {
      // first, split into smaller sub-strings
      if (plainText.length() >= PIECE_LENGTH) {
        dummy = plainText.substring(0, PIECE_LENGTH);
        plainText = plainText.substring(PIECE_LENGTH);
      }
      else {
        dummy = plainText;
        plainText = "";
      }
      // then, encrypt them individually
      dummy = encryptShort(dummy);
      // finally, concat them
      char i1 = (char)(dummy.length() % 80 + 40);
      char i2 = (char)(dummy.length() / 80 + 40);
      result += "" + i1 + i2 + dummy;
    }
    return result;
  } // end of encrypt(String)

  public String decrypt(String cipherText) {
  // the counterpart
    if (cipherText == null)
      return null;
    String result = "";
    while (cipherText.length() > 2) {
      char i1 = (char)(cipherText.charAt(0) - 40);
      char i2 = (char)(cipherText.charAt(1) - 40);
      int i = (int)i2 * 80 + (int)i1;
      cipherText = cipherText.substring(2);
      if (cipherText.length() < i)
        return null;
      String dummy = cipherText.substring(0, i);
      cipherText = cipherText.substring(i);
      result += decryptShort(dummy);
    }
    if (cipherText.length() != 0)
      return null;
    return result;
  } // end of decrypt(String)

  public String sign(String plainText) {
  // essentially, signing is no different from encrypting with the private key
    if (privateKey == null)
      return null;
    String hash = hashValue(plainText);
    return numberToString(decryptNumber(stringToNumber(encodeString(hash)))) + " " + plainText;
  }

  public String verify(String signedText) {
  // and thus, verifying is nothing other than decrypting with the public key
    if ((publicKey == null) || (signedText == null))
      return null;
    int i = signedText.indexOf(" ");
    if (i < 0)
      return null;
    String signature = signedText.substring(0, i);
    String plainText = signedText.substring(i + 1);
    String hash = decodeString(numberToString(encryptNumber(stringToNumber(signature))));
    if (hash.compareTo(hashValue(plainText)) == 0)
      return plainText;
    else
      return null;
  } // end of verify(String)

  public BigInteger getRandomPrime() {
    Random r = new Random();
    return new BigInteger(BIT_LENGTH, CERTAINTY, r);
  }

  public BigInteger getRandomComposed() {
    Random r = new Random();
    BigInteger p = new BigInteger(BIT_LENGTH/2, CERTAINTY, r);
    BigInteger q = new BigInteger(BIT_LENGTH/2, CERTAINTY, r);
    return p.multiply(q);
  }

}
